"""Tests for Jaya Clasic variant."""

import numpy as np
import math

from pyJaya.solution import Solution
from pyJaya.population import Population
from pyJaya.variants.clasic import JayaClasic
from pyJaya.variables import VariableFloat


def sphere(solution):
    return sum(np.asarray(solution)**2)


def rastrigin(solution):
    f = 0.0
    for x in solution:
        f += x ** 2 - 10 * math.cos(2 * math.pi * x) + 10
    return f


def himmelblau(solution):
    return (
        solution[0] ** 2 + solution[1] - 11) ** 2 +\
        (solution[0] + solution[1] ** 2 - 7) ** 2


def himmelblauConstraintOne(solution):
    return (26 - (solution[0] - 5) ** 2 - solution[1] ** 2) >= 0


def himmelblauConstraintTwo(solution):
    return (20 - 4 * solution[0] - solution[1]) >= 0


def truncate(number, digits):
    stepper = 10.0 ** digits
    return math.trunc(stepper * number) / stepper


class TestUnconstrainedJayaClasicInSphere:
    """Unconstrained Jaya Clasic in Sphere tests"""

    def setup(self):
        # scaling factors
        self.index = 0
        self.rn = [0.58, 0.81, 0.92, 0.49, 0.27, 0.23, 0.38, 0.51]
        # Solutions
        s1 = Solution([], sphere)
        s1.setSolution(np.array([-5.0, 18.0]))
        s2 = Solution([], sphere)
        s2.setSolution(np.array([14.0, 63.0]))
        s3 = Solution([], sphere)
        s3.setSolution(np.array([70.0, -6.0]))
        s4 = Solution([], sphere)
        s4.setSolution(np.array([-8.0, 7.0]))
        s5 = Solution([], sphere)
        s5.setSolution(np.array([-12.0, -18.0]))
        # Population
        self.population = Population(0, solutions=[s1, s2, s3, s4, s5])

    def test_first_iteration(self, monkeypatch):
        """Tests first iteration."""

        def rn():
            self.index += 1
            return self.rn[self.index-1]
        monkeypatch.setattr(np.random, 'rand', rn)

        listVars = [VariableFloat(-100.0, 100.0) for i in range(2)]
        ja = JayaClasic(5, listVars, sphere, population=self.population)
        bw = ja.run(1).getBestAndWorst()
        assert bw['best_value'] == 113.0
        assert len(bw['best_solution']) == 2
        assert all([a == b for a, b in zip(bw['best_solution'], [-8.0, 7.0])])
        assert truncate(bw['worst_value'], 1) == 3997.7
        assert len(bw['worst_solution']) == 2
        assert all(
            [a == b for a, b in zip(bw['worst_solution'], [-44.12, 45.29])])

    def test_second_iteration(self, monkeypatch):
        """Tests second iteration."""

        def rn():
            self.index += 1
            return self.rn[self.index-1]
        monkeypatch.setattr(np.random, 'rand', rn)

        listVars = [VariableFloat(-100.0, 100.0) for i in range(2)]
        ja = JayaClasic(5, listVars, sphere, population=self.population)
        bw = ja.run(2).getBestAndWorst()
        assert truncate(bw['best_value'], 2) == 7.78
        assert len(bw['best_solution']) == 2
        assert truncate(bw['best_solution'][0], 3) == 2.787
        assert truncate(bw['best_solution'][1], 3) == -0.097
        assert truncate(bw['worst_value'], 2) == 2381.13
        assert len(bw['worst_solution']) == 2
        assert truncate(bw['worst_solution'][0], 3) == -37.897
        assert truncate(bw['worst_solution'][1], 3) == 30.739


class TestUnconstrainedJayaClasicInRastringin:
    """Unconstrained Jaya Clasic in Rastringin tests"""

    def setup(self):
        self.index = 0
        # scaling factors
        self.rn = [
            0.38, 0.81, 0.92, 0.49, 0.65, 0.23,
            0.38, 0.51, 0.01, 0.7, 0.02, 0.5]
        # Solutions
        s1 = Solution([], rastrigin)
        s1.setSolution(np.array([-4.570261872, 0.045197073]))
        s2 = Solution([], rastrigin)
        s2.setSolution(np.array([3.574220009, 1.823157605]))
        s3 = Solution([], rastrigin)
        s3.setSolution(np.array([-2.304524513, 4.442417134]))
        s4 = Solution([], rastrigin)
        s4.setSolution(np.array([-1.062187325, -0.767182961]))
        s5 = Solution([], rastrigin)
        s5.setSolution(np.array([-0.84373426, 3.348170112]))
        # Population
        self.population = Population(0, solutions=[s1, s2, s3, s4, s5])

    def test_first_iteration(self, monkeypatch):
        """Tests first iteration."""

        def rn():
            self.index += 1
            return self.rn[self.index-1]
        monkeypatch.setattr(np.random, 'rand', rn)

        listVars = [VariableFloat(-5.12, 5.12) for i in range(2)]
        ja = JayaClasic(5, listVars, rastrigin, population=self.population)
        bw = ja.run(1).getBestAndWorst()
        assert truncate(bw['best_value'], 9) == 2.108371360
        assert len(bw['best_solution']) == 2
        assert truncate(bw['best_solution'][0], 9) == 0.982105143
        assert truncate(bw['best_solution'][1], 9) == -0.974135755
        assert truncate(bw['worst_value'], 9) == 36.785779376
        assert len(bw['worst_solution']) == 2
        assert truncate(bw['worst_solution'][0], 2) == 5.12
        assert truncate(bw['worst_solution'][1], 11) == -1.84339288493

    def test_second_iteration(self, monkeypatch):
        """Tests second iteration."""

        def rn():
            self.index += 1
            return self.rn[self.index-1]
        monkeypatch.setattr(np.random, 'rand', rn)

        listVars = [VariableFloat(-5.12, 5.12) for i in range(2)]
        ja = JayaClasic(5, listVars, rastrigin, population=self.population)
        bw = ja.run(2).getBestAndWorst()
        assert truncate(bw['best_value'], 9) == 2.108371360
        assert len(bw['best_solution']) == 2
        assert truncate(bw['best_solution'][0], 9) == 0.982105143
        assert truncate(bw['best_solution'][1], 9) == -0.974135755
        assert truncate(bw['worst_value'], 8) == 26.25808866
        assert len(bw['worst_solution']) == 2
        assert truncate(bw['worst_solution'][0], 10) == 2.4303683434
        assert truncate(bw['worst_solution'][1], 10) == -1.0337930258

    def test_third_iteration(self, monkeypatch):
        """Tests third iteration."""

        def rn():
            self.index += 1
            return self.rn[self.index-1]
        monkeypatch.setattr(np.random, 'rand', rn)

        listVars = [VariableFloat(-5.12, 5.12) for i in range(2)]
        ja = JayaClasic(5, listVars, rastrigin, population=self.population)
        bw = ja.run(3).getBestAndWorst()
        assert truncate(bw['best_value'], 7) == 0.2150036
        assert len(bw['best_solution']) == 2
        assert truncate(bw['best_solution'][0], 9) == -0.031679095
        assert truncate(bw['best_solution'][1], 10) == -0.0091367952
        assert truncate(bw['worst_value'], 8) == 24.79038297
        assert len(bw['worst_solution']) == 2
        assert truncate(bw['worst_solution'][0], 9) == 2.415885711
        assert truncate(bw['worst_solution'][1], 9) == -0.040158575


class TestConstrainedJayaClasicInHimmelblau:
    """Constrained Jaya Clasic in Himmelblau tests"""

    def setup(self):
        self.index = 0
        # scaling factors
        self.rn = [0.25, 0.43, 0.47, 0.33]
        # Solutions
        s1 = Solution([], himmelblau)
        s1.setSolution(np.array([3.22, 0.403]))
        s2 = Solution([], himmelblau)
        s2.setSolution(np.array([0.191, 2.289]))
        s3 = Solution([], himmelblau)
        s3.setSolution(np.array([3.182, 0.335]))
        s4 = Solution([], himmelblau)
        s4.setSolution(np.array([1.66, 4.593]))
        s5 = Solution([], himmelblau)
        s5.setSolution(np.array([2.214, 0.867]))
        # Population
        self.population = Population(0, solutions=[s1, s2, s3, s4, s5])

    def test_first_iteration(self, monkeypatch):
        """Tests first iteration."""

        def rn():
            self.index += 1
            return self.rn[self.index-1]
        monkeypatch.setattr(np.random, 'rand', rn)

        listVars = [VariableFloat(-5.0, 5.0) for i in range(2)]
        ja = JayaClasic(
            5, listVars, himmelblau, population=self.population,
            listConstraints=[
                himmelblauConstraintOne,
                himmelblauConstraintTwo])
        bw = ja.run(1).getBestAndWorst()
        assert truncate(bw['best_value'], 3) == 11.890
        assert len(bw['best_solution']) == 2
        assert truncate(bw['best_solution'][0], 3) == 3.845
        assert truncate(bw['best_solution'][1], 3) == -1.038
        assert truncate(bw['worst_value'], 3) == 77.710
        assert len(bw['worst_solution']) == 2
        assert truncate(bw['worst_solution'][0], 3) == 0.191
        assert truncate(bw['worst_solution'][1], 3) == 2.289



# -*- coding: utf-8 -*-
import numpy as np

from .base import JayaBase
from .clasic import JayaClasic
from pyJaya.population import Population
from pyJaya.solution import Solution


class JayaQuasiOppositional(JayaBase):
    """Jaya clasic class

    Args:
        numSolutions (int): Number of solutions of population.
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        space (bool): Spaced numbers over a specified interval.
        listConstraints (list, optional): Constraint list. Defaults to [].
        population (Population, optional): Population. Defaults to None.
    """

    def generateQuasiOpposite(self, population):
        """Generate quasi-opposite population

        Args:
            population (Population): Population to generate quasi-opposite.

        Returns:
            Population: [description]
        """
        newPopulation = Population(self.minimax, solutions=[])
        for solution in population.solutions:
            newSolution = Solution(
                self.listVars, self.functionToEvaluate, self.listConstraints)
            newsolution = []
            for solt_item, solt_value in enumerate(solution.solution):
                AL = self.listVars[solt_item].minor
                AU = self.listVars[solt_item].major
                a = self.listVars[solt_item].convert((AL + AU) / 2)
                b = self.listVars[solt_item].convert(AL + AU - solt_value)
                val = (max(a, b) - min(a, b)) *\
                    np.random.random_sample() + min(a, b)
                newsolution.append(self.listVars[solt_item].convert(val))
            newSolution.setSolution(newsolution)
            newPopulation.solutions.append(newSolution)
        return newPopulation

    def newPopulation(self):
        """New population with quasi-opposite elements.

        Returns:
            Population: Population with quasi-opposite elements.
        """
        auxPopulation = Population(self.minimax, solutions=[])
        quasiOppositePopulation = self.generateQuasiOpposite(self.population)
        for s in self.population.solutions:
            auxPopulation.solutions.append(s)
        for s in quasiOppositePopulation.solutions:
            auxPopulation.solutions.append(s)
        newPopulation = Population(self.minimax, solutions=[])
        if self.minimax:
            for s in auxPopulation.sorted()[-self.numSolutions:]:
                newPopulation.solutions.append(s)
        else:
            for s in auxPopulation.sorted()[:self.numSolutions]:
                newPopulation.solutions.append(s)
        return newPopulation

    def run(self, number_iterations, rn=[]):
        """Run method

        Args:
            number_iterations (int): Number of iterations.

        Returns:
            Population: Final population.
        """
        if len(rn) == 0:
            self.rn = self.generate_rn(number_iterations)
        else:
            assert number_iterations == len(rn)
            assert len(rn[0]) == self.cantVars
            assert len(rn[0][0]) == 2
            self.rn = rn
        for i in range(number_iterations):
            self.population = self.newPopulation()
            self.population = JayaClasic(
                self.population.size(), self.listVars,
                self.functionToEvaluate, listConstraints=self.listConstraints,
                population=self.population).run(1, [self.rn[i]])

        return self.population



# -*- coding: utf-8 -*-
import numpy as np

from pyJaya.variants.base import JayaBase
from pyJaya.solution import Solution


class JayaClasic(JayaBase):
    """Jaya clasic class

    Args:
        numSolutions (int): Number of solutions of population.
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        space (bool): Spaced numbers over a specified interval.
        minimaxType (minimaxType, optional): Min or Max. Defaults to [minimize]
        listConstraints (list, optional): Constraint list. Defaults to [].
        population (Population, optional): Population. Defaults to None.
    """

    def run(self, number_iterations, rn=[]):
        """Run method

        Args:
            number_iterations (int): Number of iterations.

        Returns:
            Population: Final population.
        """
        if len(rn) == 0:
            self.rn = self.generate_rn(number_iterations)
        else:
            assert number_iterations == len(rn)
            assert len(rn[0]) == self.cantVars
            assert len(rn[0][0]) == 2
            self.rn = rn
        for i in range(number_iterations):
            result = self.population.getBestAndWorst()
            for solution in self.population.solutions:
                solt = []
                for v_item, v_value in enumerate(solution.solution):
                    solt.append(self.listVars[v_item].convert(
                        (
                            v_value + self.rn[i][v_item][0] *
                            (result['best_solution'][v_item] - abs(v_value))
                            - self.rn[i][v_item][1] *
                            (result['worst_solution'][v_item]-abs(v_value)))
                    ))
                auxSolution = Solution(
                    self.listVars, self.functionToEvaluate,
                    self.listConstraints)
                auxSolution.setSolution(np.array(solt))
                if self.minimax:
                    if (auxSolution.value > solution.value) and \
                            (auxSolution.constraintsOK(np.array(solt))):
                        solution.setSolution(auxSolution.solution)
                else:
                    if (auxSolution.value < solution.value) and \
                            (auxSolution.constraintsOK(np.array(solt))):
                        solution.setSolution(auxSolution.solution)

        return self.population



# -*- coding: utf-8 -*-
"""
Solution class
"""
import numpy as np


class Solution():
    """Solution class

    Args:
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        listConstraints (list, optional): Constraint list. Defaults to [].
    """

    def __repr__(self):
        """Represent

        Returns:
            str: Solution represent.
        """
        return self.__str__()

    def __str__(self):
        """Str

        Returns:
            str: Solution represent.
        """
        return str(self.solution)

    def __init__(
            self, listVars, functionToEvaluate, listConstraints=[]):

        self.cantVars = len(listVars)
        self.listVars = listVars
        self.listConstraints = listConstraints
        self.functionToEvaluate = functionToEvaluate

        self.solution = np.zeros(self.cantVars)
        self.value = None

    def generate(self):
        """Generate a solution"""

        while True:
            solution = np.zeros(self.cantVars)
            for item in range(self.cantVars):
                solution[item] = self.listVars[item].get()
            if self.constraintsOK(solution):
                self.solution = solution
                break
        self.value = self.evaluate()

    def constraintsOK(self, solution):
        """Check that the solution satisfies the constraints

        Args:
            solution (Solution): Solution to check.

        Returns:
            Bool: True if it satisfies the constraints, False if not.
        """
        for constraints in self.listConstraints:
            if not constraints(*[solution]):
                return False
        return True

    def evaluate(self):
        """Evaluate the solution

        Returns:
            Float: Result when evaluating the solution.
        """
        return self.functionToEvaluate(tuple(*[self.solution]))

    def setSolution(self, solution):
        """Set solution

        Args:
            solution (list or np.array): Solution to be assigned.
        """
        self.solution = solution
        self.value = self.evaluate()



# -*- coding: utf-8 -*-
import numpy as np

from pyJaya.variants.base import JayaBase
from pyJaya.variants.clasic import JayaClasic
from pyJaya.population import Population


class JayaSelfAdaptive(JayaBase):
    """Jaya Self-adaptive class

    Args:
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        listConstraints (list, optional): Constraint list. Defaults to [].
    """

    def __init__(
            self, listVars, functionToEvaluate, listConstraints=[],
            population=None):
        super(JayaSelfAdaptive, self).__init__(
            len(listVars)*10, listVars, functionToEvaluate, listConstraints,
            population=population)

    def nextPopulation(self, population):
        """New population.

        Returns:
            Population: Next population.
        """
        numOldSolutions = population.size()
        r = 1 + (np.random.rand()-0.5)
        numNewSolutions = round(numOldSolutions * r)

        if numNewSolutions == numOldSolutions:
            return population
        else:
            newPopulation = Population(self.minimax, solutions=[])
            if numNewSolutions < numOldSolutions:
                if numNewSolutions < self.cantVars:
                    numNewSolutions = self.cantVars
                if self.minimax:
                    for solution in population.sorted()[-numNewSolutions:]:
                        newPopulation.solutions.append(solution)
                else:
                    for solution in population.sorted()[:numNewSolutions]:
                        newPopulation.solutions.append(solution)
            elif numNewSolutions > numOldSolutions:
                for solution in population.solutions:
                    newPopulation.solutions.append(solution)
                if self.minimax:
                    begin = numOldSolutions-numNewSolutions
                    for solution in population.sorted()[begin:]:
                        newPopulation.solutions.append(solution)
                else:
                    limit = numNewSolutions-numOldSolutions
                    for solution in population.sorted()[:limit]:
                        newPopulation.solutions.append(solution)
            return newPopulation

    def run(self, number_iterations, rn=[]):
        """Run method

        Args:
            number_iterations (int): Number of iterations.

        Returns:
            Population: Final population.
        """
        if len(rn) == 0:
            self.rn = self.generate_rn(number_iterations)
        else:
            assert number_iterations == len(rn)
            assert len(rn[0]) == self.cantVars
            assert len(rn[0][0]) == 2
            self.rn = rn
        for i in range(number_iterations):
            if i > 0:
                self.population = self.nextPopulation(self.population)
            self.population = JayaClasic(
                self.population.size(), self.listVars,
                self.functionToEvaluate, listConstraints=self.listConstraints,
                population=self.population).run(1, [self.rn[i]])

        return self.population



# -*- coding: utf-8 -*-
"""
Population class
"""
import copy
import random

import numpy as np

from pyJaya.solution import Solution
from pyJaya.consts import minimaxType


class Population():
    """Population class

    Args:
        minimax (int): Objective function to be (0: minimized or 1: maximized).
    """

    def __init__(self, minimax, solutions=[]):
        """Population init"""
        self.solutions = solutions
        self.minimax = minimax

    def generate(
            self, numSolutions, listVars, functionToEvaluate, space,
            listConstraints):
        """Population generator

        Args:
            numSolutions (int): Number of solutions.
            listVars (list): Range list.
            functionToEvaluate (funtion): Function to minimize or maximize.
            space (bool): Spaced numbers over a specified interval.
            listConstraints (list, optional): Constraint list. Defaults to [].
        """
        if space:
            v = []
            for item in listVars:
                l = np.linspace(item.minor, item.major, numSolutions).tolist()
                random.shuffle(l)
                v.append(l)
            for i in range(numSolutions):
                solution = Solution(
                    listVars, functionToEvaluate, listConstraints)
                s = []
                for item, value in enumerate(listVars):
                    s.append(value.convert(v[item][i]))
                solution.setSolution(s)
                self.solutions.append(solution)
        else:
            for i in range(numSolutions):
                solution = Solution(
                    listVars, functionToEvaluate, listConstraints)
                solution.generate()
                self.solutions.append(solution)

    def toMaximize(self):
        """Switch to maximize function
        """
        self.minimax = minimaxType['maximize']

    def sorted(self):
        """Sort the solutions

        Returns:
            list: List sorted solutions.
        """
        return sorted(
            self.solutions, reverse=self.minimax,
            key=lambda solution: solution.value)

    def getBestAndWorst(self):
        """Best and worst value and solution

        Returns:
            dict: Best value, worst value, best solution and worst solution.
        """
        solutionSorted = self.sorted()
        if self.minimax:
            return {
                'best_value': solutionSorted[-1].value,
                'worst_value': solutionSorted[0].value,
                'best_solution': solutionSorted[-1].solution,
                'worst_solution': solutionSorted[0].solution}
        else:
            return {
                'best_value': solutionSorted[0].value,
                'worst_value': solutionSorted[-1].value,
                'best_solution': solutionSorted[0].solution,
                'worst_solution': solutionSorted[-1].solution}

    def divideInTo(self, n=2):
        """Divide the population into n subpopulations

        Args:
            n (int, optional): Number of subpopulations. Defaults to 2.

        Returns:
            list: Subpopulations list.
        """
        if n == 1:
            return [self]
        populations = []
        q, r = divmod(len(self.solutions), n)
        indices = [q*i + min(i, r) for i in range(n+1)]
        subLists = [self.sorted()[indices[i]:indices[i+1]] for i in range(n)]
        for miniList in subLists:
            newPopulation = Population(self.minimax)
            newPopulation.solutions = miniList
            populations.append(newPopulation)
        return populations

    def divideInToWithElitist(self, n=2):
        """Divide the population into n subpopulations with elitism

        Args:
            n (int, optional): Number of subpopulations. Defaults to 2.

        Returns:
            list: Subpopulations list.
        """
        if n == 1:
            return [self]
        populations = []
        q, r = divmod(len(self.solutions), n)
        indices = [q*i + min(i, r) for i in range(n+1)]
        subLists = [self.sorted()[indices[i]:indices[i+1]] for i in range(n)]
        for miniList in subLists[:-1]:
            newPopulation = Population(self.minimax)
            newPopulation.solutions = miniList
            populations.append(newPopulation)
        newPopulation = Population(self.minimax)
        newPopulation.solutions = copy.deepcopy(populations[0].solutions)
        populations.append(newPopulation)
        return populations

    def merge(self, listPopulation):
        """Merge subpopulations and sort your solutions

        Args:
            listPopulation (list): Population list.
        """
        listSolutions = []
        for p in listPopulation:
            listSolutions += p.solutions
        self.solutions = sorted(
            listSolutions, reverse=self.minimax,
            key=lambda solution: solution.value)

    def size(self):
        """Get population size

        Returns:
            int: Number of solutions in the population.
        """
        return len(self.solutions)



# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

import numpy as np


class Variable(ABC):
    """Variable class
    """

    @abstractmethod
    def validate(self):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")

    @abstractmethod
    def get(self):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")

    @abstractmethod
    def convert(self):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")


class VariableInt(Variable):
    """VariableInt class

    Args:
        outside_first (int, optional):
            Lowest possible value in the range of the variable.
            Defaults to -100.
        outside_second (int, optional): Largest possible value in the
            range of the variable. Defaults to 100.
    """

    def __init__(self, outside_first=-100, outside_second=100):

        self.minor = min(outside_first, outside_second)
        self.major = max(outside_first, outside_second)
        self.validate()

    def validate(self):
        """Validate that the variables are integers

        Raises:
            NotImplementedError: "The numbers should be integer".
        """
        if not isinstance(
                self.minor, (int)) or not isinstance(self.major, (int)):
            raise NotImplementedError("The numbers should be integer")

    def get(self):
        """Returns an integer in the possible range

        Returns:
            int: Integer in the range.
        """
        np.random.seed()
        return int((self.major-self.minor) * np.random.rand() + self.minor)

    def convert(self, item):
        """Converts an item to a possible value in the range

        Args:
            item (int or float): Value to convert.

        Returns:
            int: Value in the range.
        """
        return int(np.clip(round(item), self.minor, self.major))


class VariableFloat(Variable):
    """VariableFloat class

    Args:
        outside_first (int, optional):
            Lowest possible value in the range of the variable.
            Defaults to -100.0.
        outside_second (int, optional): Largest possible value in the
            range of the variable. Defaults to 100.0.
    """

    def __init__(self, outside_first=-100.0, outside_second=100.0):

        self.minor = min(outside_first, outside_second)
        self.major = max(outside_first, outside_second)
        self.validate()

    def validate(self):
        """Validate that the variables are float

        Raises:
            NotImplementedError: "The numbers should be float".
        """
        if not isinstance(
                self.minor, (float)) or not isinstance(self.major, (float)):
            raise NotImplementedError("The numbers should be float")

    def get(self):
        """Returns an float in the possible range

        Returns:
            float: Float in the range.
        """
        np.random.seed()
        return (self.major-self.minor) * np.random.rand() + self.minor

    def convert(self, item):
        """Converts an item to a possible value in the range

        Args:
            item (int or float): Value to convert.

        Returns:
            float: Value in the range.
        """
        return float(np.clip(item, self.minor, self.major))


class VariableBinary(Variable):
    """VariableBinary class"""

    def __init__(self):

        self.minor = 0
        self.major = 1
        self.validate()

    def validate(self):
        pass

    def get(self):
        """Returns an integer in the possible range

        Returns:
            int: Integer in the range.
        """
        np.random.seed()
        return np.random.randint(2)

    def convert(self, item):
        """Converts an item to a possible value in the range

        Args:
            item (int or float): Value to convert.

        Returns:
            int: Value in the range.
        """
        return int(np.clip(round(item), self.minor, self.major))



# -*- coding: utf-8 -*-
import numpy as np
import copy

from pyJaya.consts import minimaxType
from pyJaya.population import Population


class JayaBase(object):
    """Jaya base class

    Args:
        numSolutions (int): Number of solutions of population.
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        space (bool): Spaced numbers over a specified interval.
        minimaxType (minimaxType, optional): Min or Max. Defaults to [minimize]
        listConstraints (list, optional): Constraint list. Defaults to [].
        population (Population, optional): Population. Defaults to None.
    """

    def __init__(
            self, numSolutions, listVars, functionToEvaluate, space=False,
            minimaxType=minimaxType['minimize'], listConstraints=[],
            population=None):

        super(JayaBase, self).__init__()
        self.functionToEvaluate = functionToEvaluate
        self.numSolutions = numSolutions
        self.listVars = listVars
        self.cantVars = len(listVars)
        self.minimax = minimaxType
        self.listConstraints = listConstraints
        self.space = space

        if population is None:
            self.population = self.generatePopulation()
        else:
            self.population = copy.deepcopy(population)

    def generate_rn(self, number_iterations):
        """Generate random numbers
        """
        rn = [None] * number_iterations
        for iter in range(number_iterations):
            rn[iter] = [None] * self.cantVars
            for y in range(self.cantVars):
                rn[iter][y] = [None] * 2
                for j in range(2):
                    np.random.seed()
                    rn[iter][y][j] = np.random.rand()
        return rn

    def generatePopulation(self):
        """Generate population

        Returns:
            Population: Population generated.
        """
        population = Population(self.minimax)
        population.generate(
            self.numSolutions, self.listVars, self.functionToEvaluate,
            self.space, self.listConstraints)
        return population

    def addConstraint(self, constraintFuntion):
        """Add constraint

        Args:
            constraintFuntion (funtion): Funtion to add as constraint.
        """
        self.listConstraints.append(constraintFuntion)

    def toMaximize(self):
        """Change to maximize funtion.
        """
        self.minimax = minimaxType['maximize']
        self.population.toMaximize()

    def getBestAndWorst(self):
        """Best and worst value and solution

        Returns:
            dict: Best value, worst value, best solution and worst solution.
        """
        return self.population.getBestAndWorst()

    def run(self, number_iterations, rn=[]):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")



# -*- coding: utf-8 -*-
from multiprocessing import Pool

import numpy as np

from .base import JayaBase
from .clasic import JayaClasic
from pyJaya.population import Population
from pyJaya.solution import Solution


class JayaSAMPE(JayaBase):
    """Jaya SAMPE class

    Args:
        numSolutions (int): Number of solutions of population.
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        listConstraints (list, optional): Constraint list. Defaults to [].
        population (Population, optional): Population. Defaults to None.
    """

    def sprint(self, population):
        """Jaya clasic to sub-population

        Args:
            population (Population): Population to evaluate whit Jaya clasic.

        Returns:
            Population: Sprint final population.
        """
        result = population.getBestAndWorst()
        r1 = np.random.rand(self.cantVars)
        r2 = np.random.rand(self.cantVars)
        for solution in population.solutions:
            solt = []
            for v_item, v_value in enumerate(solution.solution):
                solt.append(self.listVars[v_item].convert(
                    (
                        v_value+r1[v_item] *
                        (result['best_solution'][v_item] - abs(v_value)) -
                        r2[v_item] *
                        (result['worst_solution'][v_item]-abs(v_value)))
                ))
            auxSolution = Solution(
                self.listVars, self.functionToEvaluate,
                self.listConstraints)
            auxSolution.setSolution(np.array(solt))
            if self.minimax:
                if (auxSolution.value > solution.value) and \
                        (auxSolution.constraintsOK(np.array(solt))):
                    solution = auxSolution
            else:
                if (auxSolution.value < solution.value) and \
                        (auxSolution.constraintsOK(np.array(solt))):
                    solution.setSolution(auxSolution.solution)
        return population

    def run(self, number_iterations):
        """Run method

        Args:
            number_iterations (int): Number of iterations.

        Returns:
            Population: Final population.
        """
        result = self.population.getBestAndWorst()
        bestValue = result['best_value']
        for i in range(number_iterations):
            if i == 0:
                m = 2
                subPopulations = self.population.divideInToWithElitist(m)
                for p in subPopulations:
                    p = self.sprint(p)
                newPopulation = Population(self.minimax)
                newPopulation.merge(subPopulations)
                if self.minimax:
                    newBest = newPopulation.getBestAndWorst()['best_value']
                    lastBest = self.population.getBestAndWorst()['best_value']
                    if newBest > lastBest:
                        self.population = newPopulation
                else:
                    newBest = newPopulation.getBestAndWorst()['best_value']
                    lastBest = self.population.getBestAndWorst()['best_value']
                    if newBest < lastBest:
                        self.population = newPopulation
            else:
                if self.minimax:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV > bestValue:
                        if m < self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 1:
                        m -= 1
                    subPopulations = self.population.divideInToWithElitist(m)
                    for p in subPopulations:
                        p = self.sprint(p)
                    newPopulation = Population(self.minimax)
                    newPopulation.merge(subPopulations)
                    newBest = newPopulation.getBestAndWorst()['best_value']
                    lastBest = self.population.getBestAndWorst()['best_value']
                    if newBest > lastBest:
                        self.population = newPopulation
                else:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV < bestValue:
                        if m < self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 1:
                        m -= 1
                    subPopulations = self.population.divideInToWithElitist(m)
                    for p in subPopulations:
                        p = self.sprint(p)
                    newPopulation = Population(self.minimax)
                    newPopulation.merge(subPopulations)
                    newBest = newPopulation.getBestAndWorst()['best_value']
                    lastBest = self.population.getBestAndWorst()['best_value']
                    if newBest < lastBest:
                        self.population = newPopulation
        return self.population


class MultiprocessJayaSAMPE(JayaBase):
    """Multiprocess Jaya SAMPE class

    Args:
        numSolutions (int): Number of solutions of population.
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        listConstraints (list, optional): Constraint list. Defaults to [].
        population (Population, optional): Population. Defaults to None.
    """

    def sprint(self, population):
        """Jaya clasic to sub-population

        Args:
            population (Population): Population to evaluate whit Jaya clasic.

        Returns:
            Population: Sprint final population.
        """

        numSolutions = len(population.solutions)
        jaya_clasic = JayaClasic(
            numSolutions, self.listVars, self.functionToEvaluate,
            population=population)
        if self.minimax:
            jaya_clasic.toMaximize()

        population = jaya_clasic.run(1)
        return population

    @staticmethod
    def worker(sampe, population):
        """Worker method

        Args:
            population (Population): Original population.

        Returns:
            Population: Generated population by sprint.
        """
        return sampe.sprint(population)

    def generate(self, m):
        """Generate population

        Args:
            m (int): Number of groups based on the quality of the solutions.

        Returns:
            Population: Generated population.
        """
        entrada = self.population.divideInToWithElitist(m)
        pool = Pool(processes=3)
        results = [
            pool.apply_async(
                MultiprocessJayaSAMPE.worker,
                args=(self, x)) for x in entrada]
        subPopulations = [p.get() for p in results]

        newPopulation = Population(self.minimax)
        newPopulation.merge(subPopulations)
        self.population = newPopulation

    def run(self, number_iterations):
        """Run method

        Args:
            number_iterations (int): Number of iterations.

        Returns:
            Population: Final population.
        """
        result = self.population.getBestAndWorst()
        bestValue = result['best_value']
        m = 2
        for i in range(number_iterations):
            if i == 0:
                self.generate(2)
            else:
                if self.minimax:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV > bestValue:
                        if m < self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 2:
                        m -= 1
                else:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV < bestValue:
                        if m <= self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 2:
                        m -= 1
                self.generate(m)
        return self.population



# -*- coding: utf-8 -*-
from .base import JayaBase
from pyJaya.population import Solution
import numpy as np
import math


class JayaBinary(JayaBase):
    """Jaya binary class

    Args:
        numSolutions (int): Number of solutions of population.
        listVars (list): Range list.
        functionToEvaluate (funtion): Function to minimize or maximize.
        listConstraints (list, optional): Constraint list. Defaults to [].
        population (Population, optional): Population. Defaults to None.
    """

    def run(self, number_iterations):
        """Run method

        Args:
            number_iterations (int): Number of iterations.

        Returns:
            Population: Final population.
        """
        for i in range(number_iterations):
            result = self.population.getBestAndWorst()
            np.random.seed()
            r1 = np.random.rand(self.cantVars)
            r2 = np.random.rand(self.cantVars)
            for solution in self.population.solutions:
                solt = []
                for v_item, v_value in enumerate(solution.solution):
                    v = v_value + r1[v_item] *\
                        (result['best_solution'][v_item] - abs(v_value)) -\
                        r2[v_item] *\
                        (result['worst_solution'][v_item]-abs(v_value))
                    if math.tanh(abs(v)) > 0.5:
                        solt.append(1.0)
                    else:
                        solt.append(0.0)
                auxSolution = Solution(
                    self.listVars, self.functionToEvaluate,
                    self.listConstraints)
                auxSolution.setSolution(np.asarray(solt))
                if self.minimax:
                    if (auxSolution.value > solution.value) and \
                            (auxSolution.constraintsOK(np.asarray(solt))):
                        solution.setSolution(auxSolution.solution)
                else:
                    if (auxSolution.value < solution.value) and \
                            (auxSolution.constraintsOK(np.asarray(solt))):
                        solution.setSolution(auxSolution.solution)

        return self.population
